# Copyright 2016, 2017, 2018, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# This is various functions related to alternate paperfolding curve
# calculations, used for development and testing.  Everything likely to
# change, or move, etc.

# Most functions are named like in alternate.tex.
# Complex number returns are functions Foo_xy returning $x,$y.

package MyAlternate;
use 5.010;
use strict;
use warnings;
use Carp 'croak';
use List::Util 'min','max','sum';
use Math::BaseCnv;
use POSIX 'ceil';
use Memoize;
use Math::PlanePath::Base::Digits
  'bit_split_lowtohigh','digit_join_lowtohigh';

# uncomment this to run the ### lines
# use Smart::Comments;

use base 'Exporter';
our @EXPORT_OK = ('AlternatePaper_path',
                  'to_bigrat','to_bigint',

                  'BitAboveLowestOne','BitAboveLowestZero',
                  'turn','TurnLpred','TurnRpred',
                  'TurnLpred_flat','TurnRpred_flat',
                  'TurnRunSpred',
                  'TurnRunSpred_LtoH_flat',
                  '@TurnRunSpred_LtoH_state_table',
                  'Turn3left','Turn3right',
                  'CountOneBits','ThueMorse',
                  'Mturn',

                  'LQ','RQ','LRQ','RLQ','JA','A','AR','AL',
                  'S','D','P','Spred_k',
                  'TA',
                  'Diameter',
                  'dir', 'dir4', 'dir4_flat','@dir4_state_table',
                  'dir4_base4_flat','@dir4_base4_state_table',
                  'DirMaxN','DirMinN', 'PmOneBits','Gray',
                  'GRS','GRS_flat', 'GRSalt','GRSalt_flat',
                  'GRS4','GRS4_flat','swap23',

                  'hull_expand','hull_points_sort_by_angle','HBV',
                  '$hull6_points',

                  'point_str','points_str','points_str_gp','point_str_gp',
                  'xy_to_node_name',

                  'mul','bk','div','unfold',
                  'xy_rotate_plus90','xy_rotate_minus90','xy_rotate_180',
                  'xy_to_dir4','xy_to_dir8',
                  'rect_corners_sort',
                  'is_integer','is_pow2',

                  'PlusOffsetH_xy','PlusOffsetV_xy','PlusOffsetJ_xy',
                  'PlusToAlternate',
                  'NalternateToPlus',

                  'xy_to_tex_str','chamfer',
                  'alternate_Graph',
                  'twinalternate_Graph',
                  'twinalternate_turn_tree_Graph',

                  '@dir4_to_dx','@dir4_to_dy',
                  '@dir8_to_dx','@dir8_to_dy',

                  'TAdiameter',
                  'TAVertexToUpper','TAVertexToLower',
                  'TAVertexToLeft','TAVertexToRight',
                  'TAVertexToZ_xy',
                  'WidthS_infinity_k',
                  'TA_a','TA_c','TA_e',

                  '$Rpred_state_aref',
                  'RpredInf_flat','NonRpredInf_flat',
                  'Rpred','Rpred_of_bits',
                  '$Rpred_LtoH_state_aref',
                  'Rpred_LtoH_flat','NonRpred_LtoH_flat',

                  '$Lpred_state_aref','$Lpred_LtoH_state_aref',
                  'Lpred_k_of_bits',
                  'LpredInf_flat','NonLpredInf_flat',
                  'Lpred_k_LtoH_flat','NonLpred_k_LtoH_flat',

                  'Xnum','Gnum',
                  'Xpred_flat','Gpred_flat','SpredInf_flat','DpredInf_flat',
                  'DpredFirstInf_flat','DpredSecond_flat',
                  'Epred_flat','EpredL_flat','EpredR_flat',
                  'MfirstN','MlastN',

                  'fXpred_flat','fGpred_flat','fVpred_flat',
                  'fNonXpred_flat','fNonGpred_flat','fNonVpred_flat',

                  '$R_square_segs','$L_square_segs','$B_squares_segs',

                  '$hullsaround_segs_outer','$hullsaround_segs_inner',
                  '$hullsaround_segs_all',

                  'fBpred_hullsaround_init',
                  '@fBpred_state_table','@fBpred_state_enclosed_lists',
                  '@fBpred_hullsaround_state_to_conf',
                  '$fNonRpred_hullsaround_flat','$fNonLpred_hullsaround_flat',
                  '$fNonBpred_hullsaround_flat',
                  '$fRpred_hullsaround_flat','$fLpred_hullsaround_flat',
                  '$fBpred_hullsaround_flat',

                  'OXpred_flat',

                  'TAparentDir_flat', 'TAparentDir_LtoH_flat',
                  'TAparentDir','TAparent',
                  'TAtospineDir','TAtospine',

                  'Dean','Dean_flat','Dean_LtoH_flat',
                  '@Dean_state_table',
                  '@Dean_LtoH_state_table','@Dean_LtoH_state_to_symbol',
                  '@Dean_LtoH_state_to_name',

                  'A073057_HtoL_flat','@A073057_HtoL_state_table',
                  'A073057_LtoH_flat','@A073057_LtoH_state_table',
                  '@A073057_LtoH_state_to_name','%A073057_LtoH_name_to_state',
                 );
our %EXPORT_TAGS = (all => \@EXPORT_OK);

#------------------------------------------------------------------------------

use constant::defer AlternatePaper_path => sub {
  require Math::PlanePath::AlternatePaper;
  return  Math::PlanePath::AlternatePaper->new;
};
use constant::defer ZOrderCurve_path => sub {
  require Math::PlanePath::ZOrderCurve;
  return  Math::PlanePath::ZOrderCurve->new;
};


#------------------------------------------------------------------------------

our @dir4_to_dx = (1,0,-1,0);
our @dir4_to_dy = (0,1,0,-1);
our @dir8_to_dx = (1,1, 0,-1, -1,-1, 0,1);
our @dir8_to_dy = (0,1, 1,1,   0,-1, -1,-1);


#------------------------------------------------------------------------------
# IPC::Run bits

our @ipc;
END {
  foreach my $h (@ipc) {
    $h->finish;
  }
}

#------------------------------------------------------------------------------

# return $x,$y multiplied by 1+i
sub mul {
  my ($x,$y) = @_;
  # (x,y) + rot+90(x,y), is multiply (i+1)
  return ($x - $y,
          $y + $x);
}
sub bk {
  my ($k) = @_;
  my $y = 0 * $k;  # inherit bigint
  my $x = $y + 1;
  foreach (1 .. $k) { ($x,$y) = mul($x,$y); }
  return ($x, $y);
}

# return $x,$y multiplied by 1-i
sub div {
  my ($x,$y) = @_;
  # (x,y) + rot+90(x,y), is multiply (1-i)
  return ($x + $y,
          $y - $x);
}

# return $x,$y unfolded around pivot $ex,$ey
sub unfold {
  my ($x,$y, $ex,$ey) = @_;
  my $dx = $ex-$x;
  my $dy = $ey-$y;
  ($dx,$dy) = (-$dy,$dx);  # rotate +90
  return ($ex+$dx, $ey+$dy);
}

sub xy_rotate_plus90 {
  my ($x,$y) = @_;
  return (-$y,$x);  # rotate +90
}
sub xy_rotate_minus90 {
  my ($x,$y) = @_;
  return ($y,-$x);  # rotate -90
}
sub xy_rotate_180 {
  my ($x,$y) = @_;
  return (-$x,-$y);  # rotate 180
}

# return 0..4
sub xy_to_dir4 {
  my ($x, $y) = @_;
  require Math::NumSeq::PlanePathDelta;
  return Math::NumSeq::PlanePathDelta::_dxdy_to_dir4($x,$y);
}
# return 0..7
sub xy_to_dir8 {
  my ($x, $y) = @_;
  return 2*xy_to_dir4($x,$y);
}

sub rect_corners_sort {
  my ($x1,$y1, $x2,$y2) = @_;
  if ($x1 > $x2) {
    ($x1,$x2) = ($x2,$x1);
  }
  if ($y1 > $y2) {
    ($y1,$y2) = ($y2,$y1);
  }
  return ($x1,$y1, $x2,$y2);
}


#------------------------------------------------------------------------------

sub chamfer {
  my ($delta, $name) = @_;
  $name //= '\MyChamfer';
  if ($delta == 0) { return ''; }
  if ($delta == 1) { return "+$name"; }
  if ($delta == -1) { return "-$name"; }
  return sprintf '%+g*%s', $delta, $name;
}

sub xy_to_node_name {
  my ($x,$y) = @_;
  if ($x==int($x) && $y==int($y)) {
    if ($x < 0) { $x = '-n'.-$x; }
    if ($y < 0) { $y = 'n'.-$y; }
    return "p$x-$y";
  } else {
    return "$x,$y";
  }
}

#------------------------------------------------------------------------------
# sequences

sub LQ {
  my ($k) = @_;
  return ($k%2==0 ? 2 : 1) * 2**int($k/2) - ($k%2==0 ? 1 : 0);
}
sub RQ {
  my ($k) = @_;
  return ($k%2==0 ? 1 : 3) * 2**int($k/2) - ($k%2==0 ? 0 : 1);
}

sub LRQ {
  my ($k) = @_;
  return ($k%2==0 ? LQ($k) : RQ($k));
}
sub RLQ {
  my ($k) = @_;
  return ($k%2==0 ? RQ($k) : LQ($k));
}

sub Diameter {
  my ($k) = @_;
  if ($k <= 2) { return 2**$k; }
  return 2 * 2**int($k/2) + ($k%2==0 ? -1 : 1);
}

sub JA {
  my ($k) = @_;
  return 2**int($k/2) - 1;
}

# area, count enclosed unit squares
sub A {
  my ($k) = @_;
  return 2**($k-1) - ($k%2==0 ? 3/2 : 2)*2**int($k/2) + 1;
}

sub AR {
  my ($k) = @_;
  if ($k == 0) { return 0; }
  return 2**($k-2) - ($k%2==0 ? 1/2 : 3/2)*2**int($k/2) + ($k%2==0 ? 0 : 1);
}

sub AL {
  my ($k) = @_;
  if ($k == 0) { return 0; }
  return 2**($k-2) - ($k%2==0 ? 1 : 1/2)*2**int($k/2) + ($k%2==0 ? 1 : 0);
}

sub Gray {
  my ($n) = @_;
  return $n ^ ($n >> 1);
}
sub PmOneBits {
  my ($n) = @_;
  my $ret = 0;
  my $s = 1;
  for ( ; $n; $n>>=1) {
    $ret += ($n & 1) * $s;
    $s = -$s;
  }
  return $ret;
}

# turn at point $n, 1=left, -1=right.  Must have $n>=1.
sub turn {          # A209615
  my ($n) = @_;
  unless ($n >= 1) { croak "turn() is for n>=1"; }
  my $t = 1;
  until ($n&1) {
    $n >>= 1;
    $t = -$t;
  }
  if ($n & 2) {
    $t = -$t;
  }
  return $t;
}
sub TurnLpred {     # A106665, but OFFSET=0 there, so it is NextTurnLpred
  my ($n) = @_;
  return ($n==0 ? 0 : turn($n)==1 ? 1 : 0);
}
sub TurnRpred {     # A292077
  my ($n) = @_;
  return ($n==0 ? 0 : turn($n)==1 ? 0 : 1);
}

# turn of midpoint curve at point $n, 1=left, 0=straight, -1=right.
# Must have $n>=1.
sub Mturn {
  my ($n) = @_;
  unless ($n >= 1) { croak "Mturn() is for n>=1"; }
  my $t = turn($n);
  if ($t == turn($n+1)) {
    return $t;
  } else {
    return 0;
  }
}

sub BitAboveLowestOne {
  my ($n) = @_;
  if ($n < 1) {
    croak "BitAboveLowestOne() is for n>=1";
  }
  for (;;) {
    if (! $n || ($n % 2) != 0) {
      last;
    }
    $n = int($n/2);
  }
  $n = int($n/2);
  return ($n % 2);
}
sub BitAboveLowestZero {
  my ($n) = @_;
  for (;;) {
    if (($n % 2) == 0) {
      last;
    }
    $n = int($n/2);
  }
  $n = int($n/2);
  return ($n % 2);
}

sub Turn3left {
  my ($k) = @_;
  if ($k<=2) { return 0; }
  if ($k==3) { return 1; }
  return 2**($k-4);
}
sub Turn3right {
  my ($k) = @_;
  if ($k<=3) { return 0; }
  return 2**($k-4);
}
sub CountOneBits {
  my ($n) = @_;
  my $count = 0;
  for ( ; $n; $n>>=1) {
    $count += ($n & 1);
  }
  return $count;
}
sub ThueMorse {
  my ($n) = @_;
  return CountOneBits($n) % 2;
}

#------------------------------------------------------------------------------

sub TurnRunSpred {
  my ($n) = @_;
  $n >= 0 or croak "TurnRunSpred() is for n>=0";
  return ($n==0 ? 0
          : $n==1 || turn($n-1) != turn($n) ? 1
          : 0);
}

#                                        bit 0  1
our @TurnRunSpred_LtoH_state_table = ([ 1, 2],   #  0 = s1
                                      [ 3, 8],   #  1 = s2
                                      [ 4, 7],   #  2 = s3
                                      [ 5, 8],   #  3 = s4
                                      [ 6, 7],   #  4 = s5
                                      [ 3, 7],   #  5 = s6
                                      [ 4, 8],   #  6 = s7
                                      [ 9,10],   #  7 = s8
                                      [10, 9],   #  8 = s9
                                      [ 9, 9],   #  9 = s10
                                      [10,10]);  # 10 = s11

use constant::defer TurnRunSpred_LtoH_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA
    (\@TurnRunSpred_LtoH_state_table,
     name => "TurnRunSpred",
     accepting_list => [2,4,6,8,10]);
};


#------------------------------------------------------------------------------
# Dir

sub dir {
  my ($n) = @_;
  return PmOneBits(Gray($n));
}
sub dir4 {
  my ($n) = @_;
  return dir($n) & 3;
}

sub DirMaxN {
  my ($k) = @_;
  my $n = $k*0;  # inherit bignum class
  $k -= (~$k&1);
  for (my $i = 0; $i < $k; $i++) {
    $n <<= 1;
    if ($i%4 < 2) { $n++; }
  }
  return $n;
}
sub DirMinN {
  my ($k) = @_;
  my $n = $k*0;  # inherit bignum class
  $k -= ($k&1);
  for (my $i = 0; $i < $k; $i++) {
    $n <<= 1;
    if ($i%4 < 2) { $n++; }
  }
  return $n;
}

our @dir4_state_table = ([0,1],   # 0
                         [0,3],   # 1
                         [2,3],   # 2
                         [2,1]);  # 3

# Return a FLAT::DFA matching binary strings of n which are dir(n) = $d.
sub dir4_flat {
  my ($d) = @_;
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA
    (\@dir4_state_table,
     name => "Dir4pred d=$d",
     accepting => $d);
}
BEGIN { memoize('dir4_flat', LIST_CACHE=>'MERGE') }

our @dir4_base4_state_table = ([0,1,0,3],   # 0
                               [0,1,2,1],   # 1
                               [2,3,2,1],   # 2
                               [2,3,0,3]);  # 3

# Return a FLAT::DFA matching base4 strings of n which are dir(n) = $d.
sub dir4_base4_flat {
  my ($d) = @_;
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA
    (\@dir4_base4_state_table,
     name => "Dir4_Base4pred d=$d",
     accepting => $d);
}
BEGIN { memoize('dir4_base4_flat', LIST_CACHE=>'MERGE') }


#------------------------------------------------------------------------------
# GRS

sub GRS {
  my ($n) = @_;
  $n >= 0 or croak "GRS is for n>=0";
  my @bits = bit_split_lowtohigh($n);
  my $ret = 1;
  foreach my $i (1 .. $#bits) {
    if ($bits[$i]==1 && $bits[$i-1]==1) {
      $ret = -$ret;
    }
  }
  return $ret;
}
sub GRSalt {
  my ($n) = @_;
  if ($n & 1) { return - GRS($n); }
  else        { return   GRS($n); }
}

our %GRS_flat_accepting_dir4 = (1  => [0,1],
                                -1 => [2,3]);
sub GRS_flat {
  my ($value) = @_;
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA
    (\@dir4_state_table,
     name => sprintf('GRS %+d', $value),
     starting => 0,
     accepting_list => $GRS_flat_accepting_dir4{$value}
     || croak "GRS_flat() unrecognised value ",$value);
}
BEGIN { memoize('GRS_flat', LIST_CACHE=>'MERGE') }

our %GRSalt_flat_accepting_dir4 = (1  => [0,3],
                                   -1 => [2,1]);
sub GRSalt_flat {
  my ($value) = @_;
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA
    (\@dir4_state_table,
     name => sprintf('GRSalt %+d', $value),
     starting => 0,
     accepting_list => $GRSalt_flat_accepting_dir4{$value}
     || croak "GRSalt_flat() unrecognised value ",$value);
}
BEGIN { memoize('GRSalt_flat', LIST_CACHE=>'MERGE') }

sub swap23 {
  my ($n) = @_;
  return ($n==2 ? 3 : $n==3 ? 2 : $n);
}

sub GRS4 {
  my ($n) = @_;
  return swap23(dir4($n));
}

# Return a FLAT::DFA matching binary strings of n which are GRS4(n) = $d.
sub GRS4_flat {
  my ($d) = @_;
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA
    (\@dir4_state_table,
     name => "GRS4 pred d=$d",
     accepting => swap23($d));
}
BEGIN { memoize('GRS4_flat', LIST_CACHE=>'MERGE') }

#------------------------------------------------------------------------------
# point visits

# count of single-visited points in level $k
sub S {
  my ($k) = @_;
  return B($k)/2 + 1;
}
# count of double-visited points in level $k
sub D {
  my ($k) = @_;
  return A($k);
}

# count of total distinct points
sub P {
  my ($k) = @_;
  return S($k) + A($k);
}


#------------------------------------------------------------------------------

sub Xnum {
  my ($x) = @_;
  $x >= 0 or croak "Xnum is for x>=0";
  return Math::BaseCnv::cnv(Math::BaseCnv::cnv($x,10,2),4,10);
}
sub Gnum {
  my ($g) = @_;
  $g >= 0 or croak "Gnum is for g>=0";
  return 2*Xnum($g);
}

sub MfirstN {
  my ($s) = @_;
  $s >= 0 or croak "MfirstN is for s>=0";
  if ($s==0) { return 0; }
  Xnum(int(($s-1)/2)) + Xnum(ceil(($s-1)/2)) + 1;
}

sub MlastN {
  my ($s) = @_;
  $s >= 0 or croak "MlastN is for s>=0";
  if ($s==0) { return 0; }
  my $k = 0;
  until (2**$k <= $s && $s < 2**($k+1)) {
    $k++;
    if ($k > 20) { die "oops"; }
  }
  return 2**(2*$k+1) - Xnum(2**($k+1)-$s);
}



#------------------------------------------------------------------------------
# numbers

sub to_bigrat {
  my ($n) = @_;
  require Math::BigRat;
  return Math::BigRat->new($n);
}
sub to_bigint {
  my ($n) = @_;
  require Math::BigInt;
  return Math::BigInt->new($n);
}

sub is_integer {
  my ($n) = @_;
  return $n == int($n);
}

# return true if $n is a perfect square 1, 4, 9, etc
sub is_square {
  my ($n) = @_;
  my $s = int(sqrt($n));
  my $ss = $s*$s;
  if ($ss > $n) {
    $s--;
    $ss = $s*$s;
  }
  return $ss == $n;
}

# return true if $n is a power of 2, so 1, 2, 4, 8, 16, etc
sub is_pow2 {
  my ($n) = @_;
  (($n >> 1) < $n) or die "is_pow2() infinite";

  while ($n > 1) {
    if ($n & 1) {
      return 0;
    }
    $n >>= 1;
  }
  return ($n == 1);
}


#------------------------------------------------------------------------------
# Graph.pm
# cf Graph::Convert

# return a Graph.pm graph of alternate paperfolding curve level $k
sub alternate_Graph {
  my ($k) = @_;
  my $path = AlternatePaper_path();
  require Graph;
  my $graph = Graph->new (undirected => 1);
  $graph->set_graph_attribute(name => "Alternate $k");
  $graph->set_graph_attribute(vertex_name_type => 'xy');

  my ($n_lo, $n_hi) = $path->level_to_n_range($k);
  my ($prev_x,$prev_y) = $path->n_to_xy($n_lo);
  $graph->add_vertex("$prev_x,$prev_y");
  foreach my $n ($n_lo+1 .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    $graph->add_edge("$prev_x,$prev_y", "$x,$y");
    ($prev_x,$prev_y) = ($x,$y);
  }

  my $diameter_start = '0,0';
  my $h = 1 << ($k>>1);
  my $diameter_end = (($k&1 ? 2*$h-1 : $h) .','. ($k&1 ? 0 : $h-1) );
  ### $diameter_end
  $graph->set_graph_attribute(diameter_start => $diameter_start);
  $graph->set_graph_attribute(diameter_end   => $diameter_end);
  $graph->has_vertex($diameter_start) or die;
  $graph->has_vertex($diameter_end)   or die;
  return $graph;
}

# return a Graph.pm graph of twinalternate paperfolding curve level $k
sub twinalternate_Graph {
  my ($k) = @_;
  require Graph;
  my $graph = Graph->new (undirected => 1);
  $graph->set_graph_attribute(name => "Twin Alternate $k");
  $graph->set_graph_attribute(vertex_name_type => 'xy');

  my $path = AlternatePaper_path();
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);
  my $n_start;
  my $n_end;
  my $n_origin;
  my ($x_start,$y_start);
  if ($k%2 == 0) {
    $n_start = 3*$n_hi;
    $n_end   = 7*$n_hi;
    $n_origin = 4*$n_hi;
  } else {
    $n_start = 9*$n_hi;
    $n_end   = 13*$n_hi;
    $n_origin = 10*$n_hi;
  }
  my ($x_origin,$y_origin) = $path->n_to_xy($n_origin);
  my $transform = sub {
    my ($x,$y) = @_;
    return ($x - $x_origin, $y - $y_origin);
  };
  my ($prev_x,$prev_y) = $transform->($path->n_to_xy($n_start));
  foreach my $n ($n_start+1 .. $n_end) {
    my ($x,$y) = $transform->($path->n_to_xy($n));
    $graph->add_edge("$prev_x,$prev_y", "$x,$y");
    ($prev_x,$prev_y) = ($x,$y);
  }

  return $graph;
}

# $w vertices wide, $h vertices high
sub parallelogram_Graph {
  my ($w,$h) = @_;
  require Graph;
  my $graph = Graph->new (undirected => 1);
  $graph->set_graph_attribute(name => "Parallelogram $w,$h");
  $graph->set_graph_attribute(vertex_name_type => 'xy');
  my $add_edge_maybe = sub {
    my ($x,$y, $x2,$y2) = @_;
    if ($graph->has_vertex("$x2,$y2")) {
      $graph->add_edge("$x,$y", "$x2,$y2");
    }
  };
  my $add_vertex = sub {
    my ($x,$y) = @_;
    $graph->add_vertex("$x,$y");
    $add_edge_maybe->($x,$y, $x+1,$y);
    $add_edge_maybe->($x,$y, $x-1,$y);
    $add_edge_maybe->($x,$y, $x,$y+1);
    $add_edge_maybe->($x,$y, $x,$y-1);
  };
  foreach my $y (0 .. $h-1) {
    foreach my $x (0-$y .. $w-1-$y) {
      $add_vertex->($x,$y);
    }
  }
  return $graph;
}


#------------------------------------------------------------------------------
# Twinalternate Area Tree by Bits

sub TA_a {
  my ($k) = @_;
  return ($k==0 ? 0 : 2**($k-1));
}
sub TA_c {
  my ($k) = @_;
  return ($k==0 ? 0 : 2**($k-1) - 1);
}
sub TA_e {
  my ($k) = @_;
  return 2**$k - 1;
}

sub TAdiameter {
  my ($k) = @_;
  # vector_modulo([7,10],k) * 2^floor(k/2) - 2*k - 7
  return ($k%2==0?7:10)* 2**int($k/2) - 2*$k - 7;
}

sub TAVertexToLower {
  my ($n) = @_;
  if ($n == 0) { return undef; }
  my $mask = $n ^ ($n-1);
  return ($n & ($mask+1) ? $n - $mask - 2 : undef);
}
sub TAVertexToUpper {
  my ($n) = @_;
  my $mask = $n ^ ($n+1);
  return ($n & ($mask+1) ? undef : $n + $mask + 2);
}
sub TAVertexToLeft {
  my ($n) = @_;
  return ($n & 1 ? undef : $n ^ 1);
}
sub TAVertexToRight {
  my ($n) = @_;
  return ($n & 1 ? $n ^ 1 : undef);
}

sub TAVertexToZ_xy {
  my ($n) = @_;
  my ($x,$y) = ZOrderCurve_path()->n_to_xy($n);
  return ($y-$x, $y);
}

sub WidthS_infinity_k {
  my ($d) = @_;
  die;
  my $k = 1;
  for(;;) {
    if ($d - RQ($k-1) >=0) { $k++ }
    else { return $k-1; }
  }
}

#------------------------------------------------------------------------------
# twinalternate right turns tree

sub twinalternate_turn_tree_Graph {
  my ($k) = @_;
  my $path = AlternatePaper_path();
  my $graph = Graph->new(undirected => 1);
  $graph->set_graph_attribute (name => "Twinalternate Turn Tree $k");
  $graph->set_graph_attribute('xy',1);
  my $n_lo = 3*2**$k;
  my $n_hi = 7*2**$k;
  my $prev;
  foreach my $n ($n_lo .. $n_hi-1) {
    my $turn = ($n==$n_lo ? 1 : turn($n));
    if ($turn == -1) {
      my ($x,$y) = $path->n_to_xy($n);
      my $v = "$x,$y";
      ### $v
      $graph->add_vertex($v);
      if (defined $prev) { $graph->add_edge($v,$prev); }
      $prev = $v;
    } else {
      undef $prev;
    }
  }
  return $graph;
}


#------------------------------------------------------------------------------
# Xpred, Gpred, Rpred, Lpred

use constant::defer Xpred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new('0* (1 | []) (00|01)*')  # X axis, base-4 digits 0,1
    ->as_dfa
    ->MyFLAT::set_name('Xpred')
    ->MyFLAT::minimize;
};
use constant::defer Gpred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new('0* (00|10)*')  # X=Y diag, base-4 digits 0,2, !=0
    ->as_dfa
    ->MyFLAT::set_name('Gpred')
    ->MyFLAT::minimize;
};

sub FLAT_allow_skip_initial_0s {
  my ($fa) = @_;
  $fa = $fa->MyFLAT::as_nfa;
  my @states = $fa->epsilon_closure($fa->get_starting);
  ### starting: @states
  my %seen;
  while (@states) {
    ### seen: [keys %seen]
    @seen{@states} = ();
    @states = grep {! exists $seen{$_}}
      $fa->epsilon_closure($fa->successors(\@states,'0'));
    ### new: @states
    $fa->set_starting(@states);  # reachable from initial 0s
  }
  ### final seen: [keys %seen]
  ### starting now: [$fa->get_starting]
  return $fa;
}

# Return a FLAT::DFA which matches bit string of n which are RpredInf(n)
# right boundary segment numbers of the curve continued infinitely.
# This is simply Xpred(n) or Xpred(n+1) or Xpred(n+2).
# The bit strings of Xpred have -1 or -2 applied so as to give strings where
# n+1 or n+2 are Xpred.
#
use constant::defer RpredInf_flat => sub {
  # Xpred segment numbers -1 is those n where Xpred(n+1) is true
  require MyFLAT;
  my $fa = Xpred_flat()->union
    (Xpred_flat()->MyFLAT::digits_increment(radix => 2,
                                            add => -1),
     Xpred_flat()->MyFLAT::digits_increment(radix => 2,
                                            add => -2));
  $fa = FLAT_allow_skip_initial_0s($fa);
  return $fa->MyFLAT::minimize
    ->MyFLAT::set_name('Rpred');
};

# Return a FLAT::DFA which matches bit string of n which are NonRpredInf(n),
# ie. not right boundary segment numbers of the curve continued infinitely.
#
use constant::defer NonRpredInf_flat => sub {
  require MyFLAT;
  return RpredInf_flat()->complement
    ->MyFLAT::set_name('NonRpred')
    ->MyFLAT::minimize;
};


#------------------------------------------------------------------------------
# Dpred

# FLAT::DFA matching bits of N which are a left turn.
# Leading 0 bits are allowed.
# N=0 is not matched, neither as empty string nor as 0...
use constant::defer TurnLpred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new(' ((0|1)* 0 | []) 1 (00)*
                          | ((0|1)* 1     ) 1 0(00)*')
    ->as_dfa
    ->MyFLAT::set_name('TurnLpred')
    ->MyFLAT::minimize;
};

# FLAT::DFA matching bits of N which are a right turn.
# Leading 0 bits are allowed.
# N=0 is not matched, neither as empty string nor as 0...
use constant::defer TurnRpred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new(' ((0|1)* 0 | []) 1 0(00)*
                          | ((0|1)* 1     ) 1 (00)*')
    ->as_dfa
    ->MyFLAT::set_name('TurnRpred')
    ->MyFLAT::minimize;
};

use constant::defer SpredInf_flat => sub {
  require FLAT;
  require MyFLAT;
  return Xpred_flat()->union(Gpred_flat())
    ->MyFLAT::set_name('SpredInf')
    ->MyFLAT::minimize;
};
use constant::defer DpredInf_flat => sub {
  require FLAT;
  require MyFLAT;
  return SpredInf_flat()->complement
    ->MyFLAT::set_name('DpredInf')
    ->MyFLAT::minimize;
};

use constant::defer DpredSecond_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new('(
                            (0|1)*
                            1   1
                            (0 (0|1)) *
                            0   1
                            0*
                            |
                            ([] | (0|1)* 0)
                            1
                            (1 (0|1)) *
                            1   1
                            0*
                           )
                          ')
    ->as_dfa
    ->MyFLAT::set_name('DpredSecondInf')
    ->MyFLAT::minimize;
};
use constant::defer DpredFirstInf_flat => sub {
  require MyFLAT;
  return DpredInf_flat()->difference(DpredSecond_flat())
    ->as_dfa
    ->MyFLAT::set_name('DpredFirstInf')
    ->MyFLAT::minimize;
};


use constant::defer Epred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new('(
                            (0|1)*
                            1   1
                            (0 (0|1)) *
                            0   0
                            1*
                            |
                            ([] | (0|1)* 0)
                            1
                            (1 (0|1)) *
                            1   0
                            1*
                           )
                          ')
    ->as_dfa
    ->MyFLAT::set_name('Epred')
    ->MyFLAT::minimize;
};

use constant::defer EpredL_flat => sub {
  require MyFLAT;
  Epred_flat()->intersect(TurnLpred_flat())
    ->as_dfa
    ->MyFLAT::set_name('EpredL')
    ->MyFLAT::minimize;
};
use constant::defer EpredR_flat => sub {
  require MyFLAT;
  Epred_flat()->intersect(TurnRpred_flat())
    ->as_dfa
    ->MyFLAT::set_name('EpredR')
    ->MyFLAT::minimize;
};


#------------------------------------------------------------------------------

{
  package MyAlternateConf;
  use Carp 'croak';
  use List::Util 'min','max';
  use List::MoreUtils;

  sub new {
    my $class = shift;   # ($class, segs=>$s, side=>$side, rev=>$rev)
    if (ref $class) { $class = ref $class; }
    my $self = bless { @_ }, $class;
    $self->{'segshash'} = { map {$_=>1} $self->segs_list };
    foreach my $seg ($self->segs_list) {
      $self->seg_validate($seg);
    }
    ### $self
    return $self;
  }
  sub seg_validate {
    my ($self, $seg) = @_;
    unless ($seg =~ /^(-?[1-9]|[0-9])[0-9]*,(-?[1-9]|[0-9])[0-9]*,[0123]$/) {
      croak "bad seg ",$seg;
    }
    # x+y must be even
    my ($x,$y,$dir) = split /,/, $seg;
    if (($x+$y) & 1) {
      croak "bad seg location ",$seg;
    }
  }
  sub key {
    my ($self) = @_;
    return ($self->{'key'} //= join(' ',
                                    "side=$self->{side}",
                                    "rev=$self->{rev}",
                                    sort $self->segs_list));
  }
  sub initial {
    my $class = shift;
    return $class->new (side => 0, rev => 0, segs => ['0,0,0'], @_);
  }
  sub equals {
    my ($self, $other) = @_;
    return $self->key eq $other->key;
  }
  sub contains {
    my ($self, $seg) = @_;
    return $self->{'segshash'}->{$seg};
  }
  sub is_subset {
    my ($self, $superset) = @_;
    foreach my $seg ($self->segs_list) {
      if (! $superset->contains($seg)) { return 0; }
    }
    return 1;
  }
  sub segs_list {
    my ($self) = @_;
    return @{$self->{'segs'}};
  }
  sub num_segs {
    my ($self) = @_;
    return scalar @{$self->{'segs'}};
  }

  # return a new MyAlternateConf with each point $x,$y moved to $x+$dx, $y+$dy
  sub translate {
    my ($self, $dx,$dy) = @_;
    if ($dx==0 && $dy==0) { return $self; }
    my $name = $self->{'name'};
    if (defined $name) {
      $name = "$name moved $dx,$dy";
    }
    return $self = $self->new
      (name => $name,
       side => $self->{'side'},
       rev  => $self->{'rev'},
       segs => [ map { my ($x,$y,$dir) = split /,/;
                       $x += $dx;
                       $y += $dy;
                       "$x,$y,$dir"
                     } $self->segs_list ]);
  }

  # return a new MyAlternateConf with each segment rotated by $ddir, so each
  # $dir becomes $dir+$ddir, mod 4
  sub rotate {
    my ($self, $ddir) = @_;
    $ddir %= 4;
    if ($ddir == 0) { return $self; }

    my $name = $self->{'name'};
    if (defined $name) {
      $name = "$name rotate ".($ddir==1   ? "+90"
                               : $ddir==2 ? "180"
                               : $ddir==3 ? "-90"
                               : $ddir);
    }
    return $self = $self->new
      (name => $name,
       side => $self->{'side'},
       rev  => $self->{'rev'},
       segs => [ map { my ($x,$y,$dir) = split /,/;
                       foreach (1 .. $ddir) {
                         ($x,$y) = MyAlternate::xy_rotate_plus90($x,$y)
                       }
                       $dir = ($dir + $ddir) % 4;
                       "$x,$y,$dir"
                     } $self->segs_list ]);
  }

  # return a new MyAlternateConf with the segment common to all the given confs
  sub intersection {
    my $self = shift;
    foreach my $other (@_) {
      $self = $self->new
        (side   => $self->{'side'},
         rev    => $self->{'rev'},
         within => $self->{'within'},
         segs   => [ grep {$other->contains($_)} $self->segs_list ]);
    }
    return $self;
  }
  sub difference {
    my $self = shift;
    foreach my $other (@_) {
      $self =$self->new
        (side   => $self->{'side'},
         rev    => $self->{'rev'},
         within => $self->{'within'},
         segs   => [ grep {! $other->contains($_)} $self->segs_list ]);
    }
    return $self;
  }
  sub union {
    my $self = shift;
    my $name = $self->{'name'};
    if (defined $name) { $name = "$name union"; }
    return $self->new (name   => $name,
                       side   => $self->{'side'},
                       rev    => $self->{'rev'},
                       within => $self->{'within'},
                       segs   => [ List::MoreUtils::uniq
                                   (map {$_->segs_list} $self,@_) ]);

  }

  sub set_name {
    my ($self, $name) = @_;
    $self->{'name'} = $name;
    return $self;
  }
  sub set_within {
    my ($self, $within) = @_;
    $self->{'within'} = $within;
    return $self;
  }

  sub expand {
    my ($self, $bit) = @_;
    ($bit==0 || $bit==1) or die;
    if (! defined $self->{'side'} || ! defined $self->{'rev'}) {
      croak 'No side/rev for conf expand';
    }

    # state 0 right side enc, state 1 left side enc
    #     1^ ^0               0^ ^1
    # 0a  /   \  1b       1a  /   \  0b
    #   0/     \1           1/     \0
    #  0 ------> 1         1 ------> 0
    #  swap   same        same    swap

    my @segs = map { my ($x,$y,$dir) = split /,/;
                     $y = -$y; if ($dir & 1) { $dir ^= 2; } # conj
                     my $x2 = $x + $dir4_to_dx[$dir];
                     my $y2 = $y + $dir4_to_dy[$dir];
                     ($x,$y) = MyAlternate::mul($x,$y);
                     ($x2,$y2) = MyAlternate::mul($x2,$y2);
                     my $dir2 = ($dir-1) % 4;
                     ("$x,$y,$dir", "$x2,$y2,$dir2")
                   } $self->segs_list;
    if ($self->{'rev'} != $bit) {
      @segs = map { my ($x,$y,$dir) = split /,/;
                    $x-=1; $y-=1;
                    ($x,$y) = MyAlternate::xy_rotate_plus90($x,$y);
                    $dir = ($dir+1) % 4;
                    "$x,$y,$dir"
                  } @segs;
    }
    my $expand = $self->new (rev  => $bit,
                             side => 1-$self->{'side'},
                             segs => \@segs);
    if (defined (my $within = $self->{'within'})) {
      $expand = $expand->intersection($within);
      $expand->{'within'} = $within;
    }
    if (defined (my $name = $self->{'name'})) {
      unless ($name =~ /[0-9]$/) { $name .= ' '; }
      $expand->{'name'} = $name . '0';
    }
    return $expand;
  }

  # return true if the configuration has a boundary square side 0=right 1=left
  sub is_boundary {
    my ($self, $side) = @_;
    if (! defined $self->{'side'} || ! defined $self->{'rev'}) {
      croak 'No side,rev for conf is_boundary';
    }
    # $self->print;
    my $within = $self->{'within'} // croak 'No within';
    $side ^= $self->{'side'} ^ $self->{'rev'};
  SEG: foreach my $seg ($self->segs_list) {
      ### $seg
      my ($x,$y,$dir) = split /,/, $seg;
      my $dd = (($side ^ $dir) & 1
                ?  1  # side=1 to left, dir+1
                : -1  # side=0 to right, dir-1
               );

      #               dir
      #         x,y  ----->
      #          |          ^
      #    dir1  |          | dir3
      #          v          |
      #            <----- x2,y2
      #              dir2
      my $dir1 = ($dir + $dd) % 4;
      my $dir2 = ($dir  + 2) % 4;
      my $dir3 = ($dir1 + 2) % 4;
      my $x2 = $x + $dir4_to_dx[$dir] + $dir4_to_dx[$dir1];
      my $y2 = $y + $dir4_to_dy[$dir] + $dir4_to_dy[$dir1];

      foreach my $other_seg ("$x,$y,$dir1",
                             "$x2,$y2,$dir2",
                             "$x2,$y2,$dir3") {
        ### $other_seg
        if (! $within->contains($other_seg)) {
          # presence of $other_seg is unspecified, this square cannot be
          # used to test for boundary
          ### not within ...
          next SEG;
        }
        if (! $self->contains($other_seg)) {
          ### absent, so boundary ...
          return 1;
        }
      }
    }
    return 0;
  }

  # bounding box around the segments
  sub bbox {
    my ($self) = @_;
    my $x_min=0;
    my $x_max=1;
    my $y_min=0;
    my $y_max=0;
    foreach my $seg ($self->segs_list) {
      my ($x,$y,$dir) = split /,/, $seg;
      foreach (0,1) {
        $x_min = min($x_min, $x); $x_max = max($x_max, $x);
        $y_min = min($y_min, $y); $y_max = max($y_max, $y);
        $x += $dir4_to_dx[$dir];
        $y += $dir4_to_dy[$dir];
      }
    }
    return ($x_min,$y_min, $x_max,$y_max);
  }
  sub print {
    my ($self) = @_;
    my $segshash = $self->{'segshash'};
    print "conf ",
      $self->{'name'} // '',
      " side=", $self->{'side'}//'undef',
      " rev=", $self->{'rev'}//'undef',
      "\n";
    print "   ",join(' ',$self->segs_list),"\n";

    my ($x_min,$y_min, $x_max,$y_max) = $self->bbox;
    ### $x_min
    ### $x_max
    ### $y_min
    ### $y_max

    foreach my $y (reverse $y_min .. $y_max) {
      my $yp1 = $y+1;
      if ($y<$y_max) {
        # above verticals
        print "   ";
        foreach my $x ($x_min .. $x_max) {
          if    ($self->contains("$x,$y,1"))   { print "^     "; }
          elsif ($self->contains("$x,$yp1,3")) { print "|     "; }
          else                                 { print "      "; }
        }
        print "\n";

        print "   ";
        foreach my $x ($x_min .. $x_max) {
          if    ($self->contains("$x,$y,1"))   { print "|     "; }
          elsif ($self->contains("$x,$yp1,3")) { print "|     "; }
          else                                 { print "      "; }
        }
        print "\n";

        print "   ";
        foreach my $x ($x_min .. $x_max) {
          if    ($self->contains("$x,$y,1"))   { print "|     "; }
          elsif ($self->contains("$x,$yp1,3")) { print "v     "; }
          else                                 { print "      "; }
        }
        print "\n";
      }

      # row
      print "   ";
      foreach my $x ($x_min .. $x_max) {
        my $xp1 = $x+1;
        print $x==0&&$y==0 ? "O" : "*";
        if ($self->contains("$x,$y,0"))      { print $x==0&&$y==0
                                                 ? "====>" : "---->"; }
        elsif ($self->contains("$xp1,$y,2")) { print "<----"; }
        else                                 { print "     "; }
      }
      print "\n";
    }
    print "\n";
  }
}

#------------------------------------------------------------------------------

#    0,0 ======> 1,0
#     |           ^
#     |           |
#     v           |
#    0,-1 <----- 1,-1
#
our $R_square_segs = MyAlternateConf->new
  (name => 'R square',
   side => 0,
   rev  => 0,
   segs => [
            "0,0,0",       # below m
            "0,0,3",
            "1,-1,2",
            "1,-1,1",
           ]);

#    0,1 <------ 1,1
#     ^           |
#     |           |
#     |           v
#    0,0 ======> 1,0
#
our $L_square_segs = MyAlternateConf->new
  (name => 'L square',
   side => 0,
   rev  => 0,
   segs => [
            "0,0,0",      # above m
            "0,0,1",
            "1,1,2",
            "1,1,3",
           ]);

#    0,1 <------ 1,1
#     ^           |
#     |           |
#     |           v
#    0,0 ======> 1,0
#     |           ^
#     |           |
#     v           |
#    0,-1 <----- 1,-1
#
our $B_squares_segs = $R_square_segs->union($L_square_segs)
  ->set_name('B two squares');  # Bpred enclosing


#------------------------------------------------------------------------------
# fBpred by hulls around

our $hullsaround_segs_outer = MyAlternateConf->new
  (name => 'hullsaround outer',
   side => 0,
   rev  => 0,
   segs => [
            "-1,-1,1",
            # "-1,-1,3",
            # "0,-2,2",
            "0,-2,1",
            "1,-1,2",
            "1,-1,0",
            # "2,0,3",
            # "2,0,0",
            # "3,1,3",
            "3,1,2",
            "2,2,3",
            # "2,2,1",
            # "1,3,0",
            "1,3,3",
            "0,2,0",
            "0,2,2",
            # "-1,1,1",
            # "-1,1,2",
            # "-2,0,1",
            "-2,0,0",
           ]);

our $hullsaround_segs_inner = MyAlternateConf->new
  (name => 'hullsaround inner',
   side => 0,
   rev  => 0,
   segs => [
            "0,0,0",

            # internals
            "0,0,1",
            "0,0,2",
            "0,0,3",
            "1,1,0",
            "1,1,1",
            "1,1,2",
            "1,1,3",
            "-1,1,0",
            "-1,1,3",
            "-1,-1,0",
            "2,0,2",
            "2,0,1",
            "2,2,2",
            "1,-1,1",
            "0,2,3",
           ]);

our $hullsaround_segs_all
  = $hullsaround_segs_outer->union($hullsaround_segs_inner)->set_name('hullsaround all');

our (@fBpred_hullsaround_state_to_conf,
     @fBpred_state_table,@fBpred_state_enclosed_lists,
     $fNonRpred_hullsaround_flat,$fNonLpred_hullsaround_flat,
     $fNonBpred_hullsaround_flat,
     $fRpred_hullsaround_flat,$fLpred_hullsaround_flat,
     $fBpred_hullsaround_flat);
{
  my @state_is_RLB_enclosed;
  my @state_is_RLB_enclosed_eventually;
  @fBpred_state_enclosed_lists = ([], [], []);

  use constant::defer fBpred_hullsaround_init => sub {
    my %options = @_;
    ### fBpred_hullsaround_init() ...
    my $verbose = $options{'verbose'} || 0;

    my %conf_to_state;
    my $conf_to_state = sub {
      my ($conf) = @_;
      my $key = $conf->key;
      my $state = $conf_to_state{$key};
      if (! defined $state) {
        push @fBpred_hullsaround_state_to_conf, $conf;
        $state = $#fBpred_hullsaround_state_to_conf;
        $conf_to_state{$key} = $state;
        # print "new state $state\n";
      }
      return $state;
    };
    $conf_to_state->(MyAlternateConf->initial(within => $hullsaround_segs_all));
    ### @fBpred_hullsaround_state_to_conf
    $fBpred_hullsaround_state_to_conf[0]->is_boundary(0) or die;
    $fBpred_hullsaround_state_to_conf[0]->is_boundary(1) or die;

    {
      my $state = 0;
      my $depth = 0;
      for (;;) {
        $depth++;
        my $last_state = $#fBpred_hullsaround_state_to_conf;
        for ( ; $state <= $last_state; $state++) {
          my $from_conf = $fBpred_hullsaround_state_to_conf[$state];
          if ($verbose >= 3) {
            print "########################################\n";
            print "from $state  ",$from_conf->key,"\n";
            $from_conf->print;
          }
          foreach my $bit (0,1) {
            my $to_conf = $from_conf->expand($bit);
            my $to_state = $conf_to_state->($to_conf);
            $fBpred_state_table[$state]->[$bit] = $to_state;

            if ($verbose >= 3) {
              print "  bit=$bit to $fBpred_state_table[$state]->[$bit]  ",$to_conf->key,"\n";
              $to_conf->print;
            }
          }
        }
        my $new = $#fBpred_hullsaround_state_to_conf - $last_state;
        if ($verbose) {
          print "depth $depth new $new\n";
        }
        last if $new == 0;
      }

      # exit if $from == 6;
      # print "  $from=>{0=>$fBpred_state_table[$from]->[0],1=>$fBpred_state_table[$from]->[1]},\n";
      # {
      #   my $f = $from + 1;
      #   my $t0 = $fBpred_state_table[$from]->[0] + 1;
      #   my $t1 = $fBpred_state_table[$from]->[1] + 1;
      #   print "  $f=>{0=>$t0,1=>$t1},\n";
      # }
    }

    if ($verbose) {
      my $num_states = scalar(@fBpred_state_table);
      print "fBpred hullsaround fBpred_state_table length $num_states\n";
    }

    # print "R_squares\n";
    # foreach my $segs (@R_squares) { segs_print($segs); }
    # print "L_squares\n";
    # foreach my $segs (@L_squares) { segs_print($segs); }

    foreach my $i (0 .. $#fBpred_hullsaround_state_to_conf) {
      my $conf = $fBpred_hullsaround_state_to_conf[$i];
      foreach my $side (0, 1) {
        $state_is_RLB_enclosed[$i]->[$side] = ! $conf->is_boundary($side);
      }
      $state_is_RLB_enclosed[$i]->[2]
        = $state_is_RLB_enclosed[$i]->[0]
        && $state_is_RLB_enclosed[$i]->[1];
    }
    foreach my $side (0, 1, 2) {
      $fBpred_state_enclosed_lists[$side]
        = [ grep {$state_is_RLB_enclosed[$_]->[$side]} 0 .. $#fBpred_state_table ];
    }
    if ($verbose >= 2) {
      print "  enclosed R = ",join(',',@{$fBpred_state_enclosed_lists[0]}),
        " count ",scalar(@{$fBpred_state_enclosed_lists[0]}),"\n";
      print "  enclosed L = ",join(',',@{$fBpred_state_enclosed_lists[1]}),
        " count ",scalar(@{$fBpred_state_enclosed_lists[1]}),"\n";
      print "  enclosed B = ",join(',',@{$fBpred_state_enclosed_lists[2]}),
        " count ",scalar(@{$fBpred_state_enclosed_lists[2]}),"\n";
    }
    foreach my $side (2, 1, 0) {
      foreach my $i (@{$fBpred_state_enclosed_lists[$side]}) {
        foreach my $bit (0,1) {
          my $to = $fBpred_state_table[$i]->[$bit];
          unless ($state_is_RLB_enclosed[$to]->[$side]) {
            $fBpred_hullsaround_state_to_conf[$i]->print;
            $fBpred_hullsaround_state_to_conf[$i]->expand($bit)->print;
            print "oops, side=$side enclosed $i bit=$bit didn't remain so\n";
            # print $conf->is_boundary($side);
            exit 1;
          }
        }
      }
    }

    # # eventually enclosed configurations = those with all of $segs_eventually_B
    # {
    #   my $bad = 0;
    #   my $side = 2;
    #   foreach my $state (0 .. $#fBpred_state_table) {
    #     my $conf = $fBpred_hullsaround_state_to_conf[$state];
    #     my $enc = $state_is_RLB_enclosed_eventually[$state]->[$side];
    #     my $allof = $segs_eventually_B->is_subset($conf) ? 1 : 0;
    #     if ($enc != $allof) {
    #       print "state $state enc $enc  triples $allof\n";
    #       conf_print($conf);
    #       $bad++;
    #     }
    #   }
    #   if ($bad) { exit 1; }
    # }

    require MyFLAT;
    {
      # R
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_state_table,
         name => 'fNonRpred_hullsaround',
         accepting_list => $fBpred_state_enclosed_lists[0],
        );
      $f = MyFLAT::eventually_accepting($f);
      $fNonRpred_hullsaround_flat = $f;
      if ($verbose >= 2) {
        print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
      }

      $f = $f->complement;
      $f->{'name'} = 'fRpred_hullsaround';
      $fRpred_hullsaround_flat = $f;
    }
    {
      # L
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_state_table, name => 'fNonLpred_hullsaround',
         starting => 0,
         accepting_list => $fBpred_state_enclosed_lists[1],
        );
      $f = MyFLAT::eventually_accepting($f);
      $fNonLpred_hullsaround_flat = $f;

      $f = $f->complement;
      $f->{'name'} = 'fLpred_hullsaround';
      $fLpred_hullsaround_flat = $f;
    }
    {
      # B
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_state_table, name => 'fNonBpred_hullsaround',
         starting => 0,
         accepting_list => $fBpred_state_enclosed_lists[2],
        );
      $f = MyFLAT::eventually_accepting($f);
      $fNonBpred_hullsaround_flat = $f;
      if ($verbose >= 2) {
        print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
      }
      $f = $f->complement;
      $f->{'name'} = 'fBpred_hullsaround';
      $fBpred_hullsaround_flat = $f;
    }
  };
}

#------------------------------------------------------------------------------
# fXpred, fGpred, fVpred

{
  my @state_table = ([1,7],       # x
                     [0,5],       # g
                     [8,4],       # v
                     [6,4],       # x*
                     [2,3],       # g*
                     [1,9],       # v*
                     [6,10],      # 0
                     [10,7],      # 1
                     [10,9],      # t0
                     [8,10],      # t1
                     [10,10]);    # non
  use constant::defer fNonXpred_flat => sub {
    require MyFLAT;
    MyFLAT::aref_to_FLAT_DFA (\@state_table,
                              name => 'fNonXpred',
                              starting => 0,  # X
                              accepting => 10);
  };
  use constant::defer fNonGpred_flat => sub {
    require MyFLAT;
    MyFLAT::aref_to_FLAT_DFA (\@state_table,
                              name => 'fNonGpred',
                              starting => 1,  # G
                              accepting => 10);
  };
  use constant::defer fNonVpred_flat => sub {
    require MyFLAT;
    MyFLAT::aref_to_FLAT_DFA (\@state_table,
                              name => 'fNonVpred',
                              starting => 2,  # V
                              accepting => 10);
  };
}
use constant::defer fXpred_flat => sub {
  return fNonXpred_flat()->complement->MyFLAT::set_name('fXpred');
};
use constant::defer fGpred_flat => sub {
  return fNonGpred_flat()->complement->MyFLAT::set_name('fGpred');
};
use constant::defer fVpred_flat => sub {
  return fNonVpred_flat()->complement->MyFLAT::set_name('fVpred');
};

# fNonBpred is non-X,G,V all, so intersect
# not minimized
use constant::defer fNonBpred_flat => sub {
  return fNonXpred_flat()->intersect(fNonGpred_flat(),fNonVpred_flat())
    ->as_dfa
    ->MyFLAT::set_name('fNonBpred');
};
use constant::defer fBpred_flat => sub {
  return fNonBpred_flat()->complement->MyFLAT::set_name('fBpred');
};


#------------------------------------------------------------------------------

use constant::defer OXpred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new('0* 1 (01)* 00 0*')
    ->as_dfa
    ->MyFLAT::minimize
    ->MyFLAT::set_name('OXpred');
};


#------------------------------------------------------------------------------

our @TAparentDir_state_table = ([0,0],   # 0
                                [ 2, 1 ],   # 1=upper
                                [ 1, 4 ],   # 2=left
                                [ 3, 4 ],   # 3=lower
                                [ 2, 3 ],   # 4=right
                               );
{
  my @accepting = (4,1,2,3);

  # Return a FLAT::DFA which matches N in binary which has TAparentDir(N)==$dir.
  sub TAparentDir_flat {
    my ($dir) = @_;
    $dir %= 4;
    require MyFLAT;
    return MyFLAT::aref_to_FLAT_DFA (\@TAparentDir_state_table,
                                     name => "TAparentDir $dir",
                                     starting => 3,
                                     accepting => $accepting[$dir]);
  }
}
BEGIN { memoize('TAparentDir_flat', LIST_CACHE=>'MERGE') }

sub TAstepDir {
  my ($n,$dir) = @_;
  $dir %= 4;
  return ($dir==0   ? TAVertexToRight($n)
          : $dir==1 ? TAVertexToUpper($n)
          : $dir==2 ? TAVertexToLeft($n)
          : $dir==3 ? TAVertexToLower($n)
          : die);
}

sub TAparentDir {
  my ($n) = @_;
  $n >= 1 or croak "TAparentDir is for n>=1";
  my $state = 3;
  foreach my $bit (reverse bit_split_lowtohigh($n)) {  # high to low
    $state = $TAparentDir_state_table[$state][$bit];
  }
  return $state % 4;
}
sub TAparent {
  my ($n) = @_;
  $n >= 1 or croak "TAparent() is for n>=1";
  return TAstepDir($n, TAparentDir($n));
}

sub TAtospineDir {
  my ($n) = @_;
  $n >= 0 or croak "TAtospineDir is for n>=0";
  my $state = 1;
  foreach my $bit (reverse bit_split_lowtohigh($n)) {  # high to low
    $state = $TAparentDir_state_table[$state][$bit];
  }
  return $state % 4;
}
sub TAtospine {
  my ($n) = @_;
  $n >= 0 or croak "TAtospine() is for n>=0";
  return TAstepDir($n, TAtospineDir($n));
}

#------------------------------------------------------------------------------
# TAparentDir_LtoH

our @TAparentDir_LtoH_state_table = ([0,0],
                     [ 1, 2 ],   # s1
                     [ 7, 3 ],   # s2
                     [ 4, 2 ],   # s3
                     [ 3, 6 ],   # s4
                     [ 4, 5 ],   # s5
                     [ 6, 6 ],   # 6=yes
                     [ 7, 7 ]);  # 7=non

{
  my @starting = (2,5,4,1);

  # Return a FLAT::DFA which matches N in binary digits low to high which
  # has TAparentDir(N)==$dir.
  sub TAparentDir_LtoH_flat {
    my ($dir) = @_;
    $dir %= 4;
    my $f = MyFLAT::aref_to_FLAT_DFA (\@TAparentDir_LtoH_state_table,
                                      starting => $starting[$dir],
                                      accepting_list => [3,4,5,6]);
    if ($dir==1 || $dir==2) {
      $f = $f->complement;
    }
    $f->MyFLAT::set_name("TAparentDir $dir LtoH");
    return $f;
  }
}
BEGIN { memoize('TAparentDir_LtoH_flat', LIST_CACHE=>'MERGE') }


#------------------------------------------------------------------------------
# Richard Dean

# A003324
sub Dean {
  my ($n) = @_;
  $n >= 1 or croak "Dean_alpha() is for n>=1";
  if ($n & 1) { return $n & 3; }
  return turn($n) + 3;
}

our @Dean_state_table = ([0,1],    # s1
                         [2,3],    # s2
                         [0,3],    # s3
                         [2,1]);   # s4

# Return a FLAT::DFA which matches those n with Dean's alpha(n) == $state.
# $state is 1 to 4.
# $offset is n of the first term.  This is 1 for 
sub Dean_flat {
  my ($value, $offset) = @_;
  if (! defined $offset) { $offset = 1; }
  (1 <= $value && $value <= 4) or croak "\$value must be 1 to 4";
  my $fa = MyFLAT::aref_to_FLAT_DFA (\@Dean_state_table,
                                     starting => 0,
                                     accepting => $value-1);
  if ($offset) {
    $fa = $fa->MyFLAT::digits_increment(radix => 2,
                                        add => $offset);
  }
  return $fa->MyFLAT::set_name("Dean $value, Start=$offset");
}
BEGIN { memoize('Dean_flat', LIST_CACHE=>'MERGE') }

our @Dean_LtoH_state_table = ([1,2],   # 0 = 1a
                              [6,8],   # 1 = 1b
                              [4,3],   # 2 = 2a
                              [5,2],   # 3 = 4a
                              [7,9],   # 4 = 2b
                              [9,7],   # 5 = 4b
                              [6,6],   # 6 = 1-end
                              [7,7],   # 7 = 2-end
                              [8,8],   # 8 = 3-end
                              [9,9]);  # 9 = 4-end
our @Dean_LtoH_state_to_symbol = (1,1, 2,4,2,4, 1,2,3,4);
our @Dean_LtoH_state_to_name = ('1a','1b', '2a','4a','2b','4b',
                                '1end', '2end', '3end', '4end');
sub Dean_LtoH_flat {
  my ($value, $offset) = @_;
  (1 <= $value && $value <= 4) or croak "\$value must be 1 to 4";
  if (! defined $offset) { $offset = 1; }
  my $fa = MyFLAT::aref_to_FLAT_DFA
    (\@Dean_LtoH_state_table,
     starting => 0,
     accepting_list => [grep {$Dean_LtoH_state_to_symbol[$_] == $value}
                        0 .. $#Dean_LtoH_state_to_symbol]);
  if ($offset) {
    $fa = $fa->MyFLAT::digits_increment(radix => 2,
                                        add => $offset,
                                        direction => 'lowtohigh');
  }
  return $fa->MyFLAT::set_name("Dean $value, Start=$offset");
}
BEGIN { memoize('Dean_flat', LIST_CACHE=>'MERGE') }


#------------------------------------------------------------------------------
# A073057 concats of GRS4

# bits high to low
our @A073057_HtoL_state_table = ([0,0],  # 0 = unused
                                 [1,2],
                                 [1,3],
                                 [4,2],
                                 [4,3],
                                 [3,4],
                                 [1,5]);

# Return a FLAT::DFA which matches bit strings, with bits high to low, where
# the final state is the given $want value.
# $want is 1 to 4.
# Bit strings will be are >=2 bits, as for the "r" remainder part in A073057.
sub A073057_HtoL_flat {
  my ($want) = @_;
  return MyFLAT::aref_to_FLAT_DFA
    (\@A073057_HtoL_state_table,
     starting => 6,
     accepting => $want,
     name => "A073057 $want");
}
BEGIN { memoize('A073057_HtoL_flat', LIST_CACHE=>'MERGE') }


our @A073057_LtoH_state_to_name
  = (undef, '2a','3a','4a',
     '1a', '2b','3g','4g',  # this row is starts for low bits 0..3
     '1b', '2c','3b','4c',
     '1c', '2d','3c','4b',
     '1d', '2e','3d','4d',
     '1e', '2f','3e','4f',
     '1f',undef,'3f','4e',
     'T','U','S');
our %A073057_LtoH_name_to_state
  = (map {defined $A073057_LtoH_state_to_name[$_]
            ? ($A073057_LtoH_state_to_name[$_] => $_) : ()}
     1 .. $#A073057_LtoH_state_to_name);

our @A073057_LtoH_state_table
  = ([0,0],     # 0 = unused
     [1,2],     #   1 = 2a  -> 2a,3a
     [5,14],    #   2 = 3a  -> 2b,3c
     [8,11],    #   3 = 4a  -> 1b,4c
     [4,3],     # 4 = 1a  -> 1a,4a
     [1,9],     #   5 = 2b  -> 2a,2c
     [8,11],    #   6 = 3g  -> 1b,4c
     [18,10],   #   7 = 4g  -> 3d,3b
     [4,12],    # 8 = 1b  -> 1a,1c
     [5,14],    #   9 = 2c  -> 2b,3c
     [5,13],    #   10 = 3b  -> 2b,2d
     [19,15],   #   11 = 4c  -> 4d,4b
     [8,11],    # 12 = 1c  -> 1b,4c
     [18,10],   #   13 = 2d  -> 3d,3b
     [18,10],   #   14 = 3c  -> 3d,3b
     [8,16],    #   15 = 4b  -> 1b,1d
     [19,15],   # 16 = 1d  -> 4d,4b
     [5,13],    #   17 = 2e  -> 2b,2d
     [26,22],   #   18 = 3d  -> 3f,3e
     [23,27],   #   19 = 4d  -> 4f,4e
     [8,16],    # 20 = 1e  -> 1b,1d
     [18,17],   #   21 = 2f  -> 3d,2e
     [18,17],   #   22 = 3e  -> 3d,2e
     [23,24],   #   23 = 4f  -> 4f,1f
     [19,20],   # 24 = 1f  -> 4d,1e
     [0,0],     #   25 = unused
     [26,21],   #   26 = 3f  -> 3f,2f
     [19,20],   #   27 = 4e  -> 4d,1e
     [4,6],     # 28 = T  -> 1a,3g
     [5,7],     #   29 = U  -> 2b,4g
     [28,29]);  #   30 = S  -> T,U

# Return a FLAT::DFA which matches bit strings, with bits low to high, where
# the final state is the given $want value.
# $want is 1 to 4.
# Bit strings will be are >=2 bits, as for the "r" remainder part in A073057.
#
# $lowbits can be undef for no initial low bits, so start at "S" ready
# to take the lowest bit of r.
#
# $lowbits can be 0 to 3 for the low 2 bits of r, so r mod 4, and the DF
# thenA matches bit strings above those (bits low to high).
# The states are numbered so state $lowbits+4 is the start when such $lowbits
# are given.
#
sub A073057_LtoH_flat {
  my ($want, $lowbits) = @_;
  my $starting = 30;
  if (defined $lowbits) {
    $starting = $lowbits + 4;
  }
  # at final state, $state%4+1 is the result value
  my @accepting = grep {($_%4)+1 == $want} 1 .. 27;
  return MyFLAT::aref_to_FLAT_DFA
    (\@A073057_LtoH_state_table,
     starting => $starting,
     accepting_list => \@accepting,
     name => "A073057 LtoH $want");
}
BEGIN { memoize('A073057_LtoH_flat', LIST_CACHE=>'MERGE') }


#------------------------------------------------------------------------------

1;
__END__
