#!/usr/bin/perl -w

# Copyright 2017, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.010;
use strict;
use warnings;
use List::Util 'max','min','sum';
use Math::Geometry::Planar;
use Math::PlanePath::AlternatePaper 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyAlternate 'xy_rotate_plus90','xy_rotate_minus90','rect_corners_sort';
use MyPrintwrap ':all';
use MyPlanar;

# uncomment this to run the ### lines
# use Smart::Comments;

#
#      1.22 - |     /|
#             |    / |
#             |   |  |
#             |    \  \
#       0     +-----===-----
#                 |    |
#               0.54  1.72
#
# N to N+1
# box X to X+1, Y to Y+1

my $k = 4;
if (@ARGV) { $k = shift @ARGV; }

print "  % generated by tools/$FindBin::Script $k\n";
print "  \\newcommand{\\MyK}{$k}\n";

# libgpcl collapses sides smaller than GPC_EPSILON which by default is
# DBL_EPSILON = 1/2^53 compared to fabs(x1-x2) or some such.
# Scale up to try to avoid that.  cf DBL_MAX_EXP = 1024
# This is dependent on "double" being 64-bit IEEE style.
# Could recompile gpcl with GPC_EPSILON=0 instead if preferred.
#
my $scale = 2**64;

my $path = Math::PlanePath::AlternatePaper->new;
my $n_plus_1_lo = max(2, 4**$k);
my $n_plus_1_hi = 4**($k+1)-1;  # inclusive
print "  % n+1= $n_plus_1_lo to $n_plus_1_hi inclusive\n";

# accumulated contour
my $gpc = MyPlanar::points_to_gpc([ [0,0], [0,0], [0,0] ]);

my $x_min;
my $x_max;
my $y_min;
my $y_max;

# point(n+1)/sqrt(n)    =>  point(n)/sqrt(n-1)
# point(n+1)/sqrt(n+1)      point(n)/sqrt(n)

sub n_plus_1_to_polygon {
  my ($n_plus_1) = @_;
  my ($x1,$y1) = $path->n_to_xy($n_plus_1);
  my ($dx,$dy) = $path->n_to_dxdy($n_plus_1);
  my $x2 = $x1 + $dx;
  my $y2 = $y1 + $dy;

  ### n_to_polygon(): "n+1=$n_plus_1  xy=$x1,$y1  $dx,$dy to $x2,$y2"

  $y1 <= $x1 or die;
  $y2 <= $x2 or die;

  my $n = $n_plus_1 - 1;
  if ($n <= 0) { return undef; }
  my $sqrt_big   = $scale / sqrt($n);
  my $sqrt_small = $scale / sqrt($n_plus_1);

  my $x1_small = $x1 * $sqrt_small;
  my $y1_small = $y1 * $sqrt_small;
  my $x1_big = $x1 * $sqrt_big;
  my $y1_big = $y1 * $sqrt_big;

  my $x2_small = $x2 * $sqrt_small;
  my $y2_small = $y2 * $sqrt_small;
  my $x2_big = $x2 * $sqrt_big;
  my $y2_big = $y2 * $sqrt_big;

  $y1_big <= $x1_big or die;
  $y2_big <= $x2_big or die;

  my ($x3,$y3);

  if ($dx) {
    # even, triangle on left
    #
    #                 *  lx,ly          x2,y2  *<==*  x1,y1
    #                /|                        |  /
    #               / |                        | /
    #              /  |                        |/
    #      x1,y1  *==>*  x2,y2          lx,ly  *
    #
    my ($lx,$ly) = xy_rotate_plus90 ($dx,$dy); # left
    $x3 = $x2 + $lx;
    $y3 = $y2 + $ly;

  } elsif ($dy) {
    # odd, triangle on right
    #
    #     x2,y1  *              rx,ry  *---*  x1,y1
    #            ^\                     \  |       
    #            | \                     \ |       
    #            |  \                     \v       
    #     x1,y1  *---* rx,ry               *  x2,y2
    #
    my ($rx,$ry) = xy_rotate_minus90 ($dx,$dy); # right
    $x3 = $x1 + $rx;
    $y3 = $y1 + $ry;
  } else {
    die;
  }

  my $points = [ [ $x1*$sqrt_big,   $y1*$sqrt_big   ],
                 [ $x2*$sqrt_big,   $y2*$sqrt_big   ],
                 [ $x3*$sqrt_big,   $y3*$sqrt_big   ],
                 [ $x1*$sqrt_small,   $y1*$sqrt_small   ],
                 [ $x2*$sqrt_small,   $y2*$sqrt_small   ],
                 [ $x3*$sqrt_small,   $y3*$sqrt_small   ],
               ];
  my $planar = Math::Geometry::Planar->new;
  $planar->points($points);
  $planar = $planar->convexhull2;

  return $planar->points;
}

foreach my $n_plus_1 ($n_plus_1_lo .. $n_plus_1_hi) {
  ### $n_plus_1

  my $points = n_plus_1_to_polygon ($n_plus_1);
  ### $points
  next unless defined $points;

  # print_poly($points);
  # next;

  $x_min = min($x_min//(), map {$_->[0]} @$points);
  $x_max = max($x_max//(), map {$_->[0]} @$points);
  $y_min = min($y_min//(), map {$_->[1]} @$points);
  $y_max = max($y_max//(), map {$_->[1]} @$points);

  # if ($y1==0||$y2==0||$y3==0) {
  #   print "at $n_plus_1  xy=",join(',',$path->n_to_xy($n_plus_1)),"\n";
  # }

  my $gpc_subcurve = MyPlanar::points_to_gpc($points);
  $gpc = Math::Geometry::Planar::GpcClip ('UNION', $gpc, $gpc_subcurve);
  if (! defined $gpc) {
    die;
  }
}
$x_min /= $scale;
$x_max /= $scale;
$y_min /= $scale;
$y_max /= $scale;

sub coordinate_to_str {
  my ($x) = @_;
  $x = sprintf '%.8f', $x;
  $x =~ s/0+$//;
  $x =~ s/([0-9])\.$/$1/;
  $x =~ s/^0\././;
  return $x;
}

sub point_Manhattan {
  my ($p) = @_;
  return sum(@$p);
}
sub point_Leading {
  my ($p) = @_;
  return $p->[0]-$p->[1];
}
sub scaled_printwrap_poly {
  my ($points) = @_;
  return unless @$points;
  my $join = ' ';
  foreach my $p (@$points) {
    my ($x,$y) = @$p;
    $x /= $scale;
    $y /= $scale;
    my $x_str = coordinate_to_str($x);
    my $y_str = coordinate_to_str($y);
    printwrap("$join($x_str,$y_str)");
    $join = '--';
  }
  printwrap("${join}cycle");
}

my @points_list;
my @contours = Math::Geometry::Planar::Gpc2Polygons($gpc);
my $total_holes = 0;
foreach my $contour (@contours) {  # Math::Geometry::Planar contour
  ### $contour
  # Second and subsequent polygon of the contour are holes.
  # Think their points already cycle the opposite way as required by pgf
  # default \fill[nonzero rule], but using \fill[even odd rule] works either
  # way.
  $Printwrap = 2;
  printwrap(" \\fill[even odd rule]");
  $Printwrap += 2;
  my $polygons = $contour->polygons;  # arrayref of points arrayrefs
  $total_holes += scalar(@$polygons)-1;
  my $separator = '';
  foreach my $points (@$polygons) {
    ### $points
    printwrap($separator);
    $separator = "\n";
    scaled_printwrap_poly($points);
    push @points_list, @$points;
  }
  printwrap(";\n");
}

my $num_contours = scalar(@contours);
my $total_points = scalar(@points_list);

my $points_x_min = min(map {$_->[0]} @points_list) / $scale;
my $points_x_max = max(map {$_->[0]} @points_list) / $scale;
my $points_y_min = min(map {$_->[1]} @points_list) / $scale;
my $points_y_max = max(map {$_->[1]} @points_list) / $scale;
my $points_m_min = min(map {point_Manhattan($_)} @points_list) / $scale;
my $points_m_max = max(map {point_Manhattan($_)} @points_list) / $scale;
my $points_l_min = min(map {point_Leading($_)}   @points_list) / $scale;
my $points_l_max = max(map {point_Leading($_)}   @points_list) / $scale;
# ### @points_list

print "  % $num_contours contours, $total_holes holes, $total_points total points\n";
printf "  %% smallest side %.5f\n", 1/sqrt($n_plus_1_hi);
printf "  %% parts X %.5f %.5f  Y %.5f %.5f\n", $x_min, $x_max, $y_min, $y_max;
printf "  %% final X %.5f %.5f  Y %.5f %.5f\n",
  $points_x_min, $points_x_max, $points_y_min, $points_y_max;
printf "  %%       M %.5f %.5f  L %.5f %.5f\n",
  $points_m_min, $points_m_max, $points_l_min, $points_l_max;
print "  % end generated\n";
exit 0;
